import asyncio
import dataclasses
import functools
import pathlib
import re
import shlex
import shutil
import subprocess
import sys
from enum import Enum
from typing import IO, Any, Dict, Iterable, List, Optional, Tuple, Union

import typer
from pydantic import BaseModel, Field
from rich.text import Text

from rbx import utils
from rbx.box import safeeval
from rbx.box.exception import RbxException
from rbx.config import get_bits_stdcpp, get_jngen, get_testlib
from rbx.console import console
from rbx.grading import grading_context
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.program import ProgramError
from rbx.grading.judge.sandbox import (
    CommunicationParams,
    SandboxBase,
    SandboxLog,
    SandboxParams,
)
from rbx.grading.judge.storage import copyfileobj
from rbx.grading.limits import Limits

MAX_STDOUT_LEN = 1024 * 1024 * 128  # 128 MB


class Outcome(Enum):
    ACCEPTED = 'accepted'
    WRONG_ANSWER = 'wrong-answer'
    MEMORY_LIMIT_EXCEEDED = 'memory-limit-exceeded'
    TIME_LIMIT_EXCEEDED = 'time-limit-exceeded'
    IDLENESS_LIMIT_EXCEEDED = 'idleness-limit-exceeded'
    RUNTIME_ERROR = 'runtime-error'
    OUTPUT_LIMIT_EXCEEDED = 'output-limit-exceeded'
    JUDGE_FAILED = 'judge-failed'
    INTERNAL_ERROR = 'internal-error'
    COMPILATION_ERROR = 'compilation-error'

    @classmethod
    def worst_outcome(cls, outcomes: Iterable['Outcome']) -> 'Outcome':
        def _outcome_to_int(o: 'Outcome') -> int:
            return cls._member_names_.index(o.name)

        return max(outcomes, key=_outcome_to_int)

    def is_slow(self) -> bool:
        return self in [
            Outcome.TIME_LIMIT_EXCEEDED,
            Outcome.IDLENESS_LIMIT_EXCEEDED,
        ]

    def is_limit_exceeded(self) -> bool:
        return self in [
            Outcome.TIME_LIMIT_EXCEEDED,
            Outcome.MEMORY_LIMIT_EXCEEDED,
            Outcome.OUTPUT_LIMIT_EXCEEDED,
            Outcome.IDLENESS_LIMIT_EXCEEDED,
        ]

    def short_name(self) -> str:
        if self == Outcome.ACCEPTED:
            return 'AC'
        if self == Outcome.WRONG_ANSWER:
            return 'WA'
        if self == Outcome.TIME_LIMIT_EXCEEDED:
            return 'TLE'
        if self == Outcome.IDLENESS_LIMIT_EXCEEDED:
            return 'ILE'
        if self == Outcome.MEMORY_LIMIT_EXCEEDED:
            return 'MLE'
        if self == Outcome.RUNTIME_ERROR:
            return 'RTE'
        if self == Outcome.OUTPUT_LIMIT_EXCEEDED:
            return 'OLE'
        if self == Outcome.JUDGE_FAILED:
            return 'FL'
        if self == Outcome.INTERNAL_ERROR:
            return 'IE'
        if self == Outcome.COMPILATION_ERROR:
            return 'CE'
        return 'XX'


class DigestHolder(BaseModel):
    value: Optional[str] = None


class GradingLogsHolder(BaseModel):
    run: Optional['RunLog'] = None
    interactor_run: Optional['RunLog'] = None
    preprocess: Optional[List['PreprocessLog']] = None
    cached: bool = False


class DigestOrSource(BaseModel):
    # Source path relative to the FS.
    src: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None

    @staticmethod
    def create(data: Union[pathlib.Path, DigestHolder, str]) -> 'DigestOrSource':
        if isinstance(data, str):
            return DigestOrSource(digest=DigestHolder(value=data))
        if isinstance(data, DigestHolder):
            return DigestOrSource(digest=data)
        return DigestOrSource(src=data)

    def expand(self) -> Dict[str, Any]:
        res = {}
        if self.src is not None:
            res['src'] = self.src
        if self.digest is not None:
            res['digest'] = self.digest
        return res


class DigestOrDest(BaseModel):
    # Destination path relative to the FS.
    dest: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None

    @staticmethod
    def create(data: Union[pathlib.Path, DigestHolder, str]) -> 'DigestOrDest':
        if isinstance(data, str):
            return DigestOrDest(digest=DigestHolder(value=data))
        if isinstance(data, DigestHolder):
            return DigestOrDest(digest=data)
        return DigestOrDest(dest=data)

    def expand(self) -> Dict[str, Any]:
        res = {}
        if self.dest is not None:
            res['dest'] = self.dest
        if self.digest is not None:
            res['digest'] = self.digest
        return res


class GradingFileInput(BaseModel):
    # Destination path relative to the sandboox.
    dest: pathlib.Path
    # Source path relative to the FS.
    src: Optional[pathlib.Path] = None
    # Digest if we should get file from storage.
    digest: Optional[DigestHolder] = None
    # Whether the destination file should be marked as an executable.
    executable: bool = False
    # Whether to track file through its hash (disable for optimization).
    hash: bool = True


class GradingFileOutput(BaseModel):
    # Source path relative to the sandbox.
    src: pathlib.Path
    # Destination path relative to the FS.
    dest: Optional[pathlib.Path] = None
    # Digest if we should put file in storage.
    digest: Optional[DigestHolder] = None
    # Whether the destination file should be marked as an executable.
    executable: bool = False
    # Whether the file is optional or not.
    optional: bool = False
    # Whether to cap its size
    maxlen: Optional[int] = None
    # Whether the file is just an intermediate file that should not be tracked.
    intermediate: bool = False
    # Whether to track file through its hash (disable for optimization).
    hash: bool = True
    # Whether to touch the file before the command runs.
    touch: bool = False

    def get_file(self, cacher: FileCacher) -> Optional[IO[bytes]]:
        if self.dest is not None:
            if self.optional and not self.dest.exists():
                return None
            return self.dest.open('rb')
        if self.digest is not None and self.digest.value is not None:
            if self.optional and not cacher.exists(self.digest.value):
                return None
            return cacher.get_file(self.digest.value)
        raise ValueError('No file to get')


class GradingFifo(BaseModel):
    # Destination path relative to the sandbox.
    path: pathlib.Path
    # Symlink to the FIFO outside the sandbox.
    symlink: Optional[pathlib.Path] = None
    # Whether to create the FIFO if it does not exist.
    create: bool = True


class GradingArtifacts(BaseModel):
    # Root directory for the produced artifacts.
    root: pathlib.Path = pathlib.PosixPath('.')
    # List of input files to copy to the sandbox.
    inputs: List[GradingFileInput] = []
    # List of output files to copy from the sandbox.
    outputs: List[GradingFileOutput] = []
    # List of FIFOs
    fifos: List[GradingFifo] = []
    # Capture certain logs of the execution.
    logs: Optional[GradingLogsHolder] = None

    def get_input_file_for_dest(self, dest: pathlib.Path) -> Optional[GradingFileInput]:
        for input in self.inputs:
            if input.dest == dest:
                return input
        return None

    def get_output_file_for_src(self, src: pathlib.Path) -> Optional[GradingFileOutput]:
        for output in self.outputs:
            if output.src == src:
                return output
        return None


class TestcaseIO(BaseModel):
    index: int
    input: Optional[pathlib.Path] = None
    output: Optional[pathlib.Path] = None


class RunLogMetadata(BaseModel):
    language: Optional[str] = None
    is_sanitized: bool = False
    limits: Limits = Field(default_factory=Limits)
    timeLimit: Optional[int] = None
    memoryLimit: Optional[int] = None
    retryIndex: Optional[int] = None


class ProcessingContextLog(BaseModel):
    pid: int = -1
    exitindex: int = -1


class RunLog(BaseModel):
    exitcode: int = 0
    exitstatus: str = SandboxBase.EXIT_SANDBOX_ERROR
    time: Optional[float] = 0.0
    memory: Optional[int] = 0
    sandbox: str = ''
    warnings: bool = False
    metadata: Optional[RunLogMetadata] = None
    exitindex: int = 0

    def get_run_language(self) -> Optional[str]:
        if self.metadata is None:
            return None
        return self.metadata.language

    def get_summary(self) -> str:
        if self.exitcode == 0:
            return 'OK'
        time = self.time or 0.0
        memory = self.memory or 0
        return f'FAILED with exit code {self.exitcode} and sandbox status {self.exitstatus} (time: {time}s, memory: {memory // (1024 * 1024)}MB)'


class PreprocessLog(RunLog):
    cmd: List[str]
    log: str

    def get_command(self) -> str:
        return ' '.join(self.cmd)


class TestcaseLog(RunLog):
    stdout_absolute_path: Optional[pathlib.Path] = None
    stderr_absolute_path: Optional[pathlib.Path] = None
    log_absolute_path: Optional[pathlib.Path] = None
    eval_absolute_path: Optional[pathlib.Path] = None


class CheckerResult(BaseModel):
    outcome: Outcome
    message: str = ''
    no_tle_outcome: Optional[Outcome] = None
    sanitizer_warnings: bool = False


class Evaluation(BaseModel):
    result: CheckerResult
    testcase: TestcaseIO
    log: TestcaseLog


def _process_input_artifacts(artifacts: GradingArtifacts, sandbox: SandboxBase):
    for input_artifact in artifacts.inputs:
        if input_artifact.digest is not None:
            assert input_artifact.digest.value is not None
            sandbox.create_file_from_storage(
                input_artifact.dest,
                input_artifact.digest.value,
                override=True,
                executable=input_artifact.executable,
                try_symlink=True,
            )
            continue
        assert input_artifact.src is not None
        sandbox.create_file_from_other_file(
            input_artifact.dest,
            artifacts.root / input_artifact.src,
            executable=input_artifact.executable,
            override=True,
            try_symlink=True,
        )
    for output_artifact in artifacts.outputs:
        if output_artifact.touch:
            sandbox.create_file_from_string(
                output_artifact.src,
                '',
                executable=output_artifact.executable,
                override=True,
            )


def _process_output_artifacts(
    artifacts: GradingArtifacts,
    sandbox: SandboxBase,
) -> bool:
    for output_artifact in artifacts.outputs:
        if output_artifact.hash and output_artifact.digest is None:
            if not grading_context.is_no_cache():
                # If cache is enabled, track this file in cache.
                output_artifact.digest = DigestHolder()
        if not sandbox.file_exists(output_artifact.src):
            if output_artifact.optional:
                continue
            console.print(
                f'[error]Output artifact [item]{output_artifact.src}[/item] does not exist.[/error]'
            )
            return False

        if output_artifact.digest is not None:
            # Put it in the cache, possibly compressing it if it's an executable.
            with grading_context.compression(
                use_compression=True,
                when=output_artifact.executable,
            ):
                output_artifact.digest.value = sandbox.get_file_to_storage(
                    output_artifact.src, trunc_len=output_artifact.maxlen
                )
        if output_artifact.dest is None:
            continue
        dst: pathlib.Path = artifacts.root / output_artifact.dest
        # Ensure dst directory exists.

        dst.parent.mkdir(parents=True, exist_ok=True)

        if (
            output_artifact.digest is not None
            and output_artifact.digest.value is not None
            and (
                path_to_symlink := sandbox.file_cacher.path_for_symlink(
                    output_artifact.digest.value
                )
            )
            is not None
        ):
            # File is in the persistent cache, store a symlink to it.
            dst.unlink(missing_ok=True)
            dst.parent.mkdir(parents=True, exist_ok=True)
            dst.symlink_to(path_to_symlink)
        else:
            # File is not in the persistent cache, copy it.
            with dst.open('wb') as f:
                with sandbox.get_file(output_artifact.src) as sb_f:
                    copyfileobj(
                        sb_f,
                        f,
                        maxlen=output_artifact.maxlen,
                    )
        if output_artifact.executable:
            dst.chmod(0o755)
    return True


def _process_fifos(artifacts: GradingArtifacts, sandbox: SandboxBase):
    for fifo in artifacts.fifos:
        if fifo.symlink is not None:
            sandbox.create_symlink(fifo.path, fifo.symlink, override=True)
        else:
            sandbox.create_fifo(fifo.path, override=True)


def testlib_grading_input() -> GradingFileInput:
    return GradingFileInput(src=get_testlib(), dest=pathlib.Path('testlib.h'))


def jngen_grading_input() -> GradingFileInput:
    return GradingFileInput(src=get_jngen(), dest=pathlib.Path('jngen.h'))


def _expand_part(part: str, sandbox: SandboxBase) -> List[str]:
    part = part.strip()
    if part.startswith('@glob:'):
        return [str(path) for path in sandbox.glob(part[6:])]
    return [part]


def _get_java_memory_limits(params: SandboxParams) -> Tuple[int, int]:
    max_memory = params.address_space
    if max_memory is None:
        max_memory = 2048
    return max_memory, min(512, int(max_memory * 0.9))


def _split_and_expand(
    command: str, sandbox: SandboxBase, params: SandboxParams
) -> List[str]:
    res = []
    max_mem, init_mem = _get_java_memory_limits(params)
    command = safeeval.eval_as_fstring(
        command, {'memory': max_mem, 'initialMemory': init_mem}
    )
    parts = shlex.split(command)
    for part in parts:
        res.extend(_expand_part(part, sandbox))
    return res


def get_exe_from_command(command: str) -> str:
    cmds = shlex.split(command)
    if not cmds:
        return command
    return cmds[0]


def _is_c_command(exe_command: str) -> bool:
    return 'gcc' in exe_command or 'clang' in exe_command


def is_cpp_command(exe_command: str) -> bool:
    return 'g++' in exe_command or 'clang++' in exe_command


def is_cxx_command(exe_command: str) -> bool:
    return is_cpp_command(exe_command) or _is_c_command(exe_command)


def is_cxx_sanitizer_command(command: str) -> bool:
    exe = get_exe_from_command(command)
    if not exe:
        return False
    if not is_cxx_command(exe):
        return False
    return 'fsanitize' in command


def is_java_command(exe_command: str) -> bool:
    return 'javac' in exe_command or 'java' in exe_command


def is_kotlin_command(exe_command: str) -> bool:
    return 'kotlinc' in exe_command or 'kotlin' in exe_command


def is_java_like_command(exe_command: str) -> bool:
    return is_java_command(exe_command) or is_kotlin_command(exe_command)


@functools.cache
def _complain_about_clang() -> None:
    console.print(
        '[warning]Notice your C++ files are being compiled with [item]clang[/item] instead of [item]g++[/item].[/warning]'
    )
    console.print('[warning]This may lead to unexpected behavior.[/warning]')
    console.print('[warning]Consider using [item]g++[/item] instead.[/warning]')
    console.print(
        '[warning]See [item]https://rsalesc.github.io/rbx/cpp-on-macos[/item] for instructions on how to use [item]g++[/item] on MacOS.'
    )


@functools.cache
def _get_cxx_version_output(command: str, extra_flags: str = '') -> Optional[str]:
    cmds = shlex.split(command)
    if not cmds:
        return None
    exe = cmds[0]
    if not is_cxx_command(exe):
        return None

    extra = shlex.split(extra_flags)
    output = subprocess.run([exe, '-v', *extra], capture_output=True, input='')
    if output.returncode != 0:
        console.print('[error]Failed to get C/C++ compiler version.[/error]')
        return None
    return output.stderr.decode()


def _maybe_get_bits_stdcpp_for_clang(command: str) -> Optional[GradingFileInput]:
    if not is_cpp_command(get_exe_from_command(command)):
        return None
    version_output = _get_cxx_version_output(command)
    if version_output is None:
        return None
    lines = version_output.splitlines()
    if not lines:
        return None
    # Check the first line for `clang`.
    if 'clang' not in lines[0]:
        return None

    if not is_cxx_sanitizer_command(command):
        _complain_about_clang()
    bits = get_bits_stdcpp()
    return GradingFileInput(src=bits, dest=pathlib.Path('bits/stdc++.h'))


def _find_system_paths_in_version_output(version_output: str) -> List[pathlib.Path]:
    res = []
    start = False
    for line in version_output.splitlines():
        if line.startswith('#include <...> search starts here:'):
            start = True
            continue
        if not start:
            continue
        if not line.startswith(' '):
            break
        res.append(pathlib.Path(line.strip()))
    return res


def _get_system_bits_stdcpp(command: str) -> Optional[GradingFileInput]:
    if not is_cpp_command(get_exe_from_command(command)):
        return None
    version_output = _get_cxx_version_output(command, '-xc++ -E -')
    if version_output is None:
        return None
    for path in _find_system_paths_in_version_output(version_output):
        bits_candidate = path / 'bits' / 'stdc++.h'
        if not bits_candidate.is_file():
            continue
        return GradingFileInput(
            src=utils.abspath(bits_candidate),
            dest=pathlib.Path('bits/stdc++.h'),
        )
    return None


def maybe_get_bits_stdcpp_for_commands(
    commands: List[str],
) -> Optional[GradingFileInput]:
    for command in commands:
        res = _get_system_bits_stdcpp(command) or _maybe_get_bits_stdcpp_for_clang(
            command
        )
        if res is not None:
            return res
    return None


@functools.cache
def _try_following_alias_for_exe(exe: str) -> Optional[str]:
    output = subprocess.run(
        f'which {exe}', shell=True, executable=shutil.which('bash'), capture_output=True
    )
    if output.returncode != 0:
        return None
    return output.stdout.decode().strip()


def _try_following_alias_for_command(command: str) -> str:
    cmds = shlex.split(command)
    if not cmds:
        return command
    exe = cmds[0]
    new_exe = _try_following_alias_for_exe(exe)
    if new_exe is None:
        return command
    return shlex.join([new_exe, *cmds[1:]])


def _try_following_alias_for_commands(commands: List[str]) -> List[str]:
    res: List[str] = []
    for command in commands:
        res.append(_try_following_alias_for_command(command))
    return res


@functools.cache
def _maybe_complain_about_sanitization(command: str) -> None:
    if not is_cxx_sanitizer_command(command):
        return
    if sys.platform != 'darwin':
        return

    version_output = _get_cxx_version_output(command)
    if version_output is None:
        return
    lines = version_output.splitlines()
    if not lines:
        return
    if 'gcc' in lines[-1]:
        console.print(
            '[error]Notice you are using sanitizers in [item]MacOS[/item], but your C/C++ compiler is [item]gcc[/item].[/error]'
        )
        console.print('[error]GCC does not support sanitization in MacOS.[/error]')
        console.print(
            '[warning]See [item]https://rsalesc.github.io/rbx/cpp-on-macos[/item] for instructions on how to use C/C++ sanitizers on MacOS.[/warning]'
        )
        raise typer.Exit(1)


def check_for_sanitizer_warnings_in_line(line: str) -> bool:
    line = line.lower()
    return 'runtime error:' in line or '==error' in line


def _check_for_sanitizer_warnings(
    sandbox: SandboxBase, stderr_file: Optional[pathlib.Path]
) -> bool:
    if stderr_file is None:
        return False
    if not sandbox.file_exists(stderr_file):
        return False
    with sandbox.get_file(stderr_file) as f:
        return any(check_for_sanitizer_warnings_in_line(line.decode()) for line in f)


_WARNING_RE = re.compile(r'([^:]+):\d+:\d+:[ ]+warning:.*')


def _check_for_compilation_warnings_in_line(line: str) -> bool:
    if line.startswith('./'):
        return False
    line = utils.strip_ansi_codes(line)
    match = _WARNING_RE.match(line)
    if match is None:
        return False
    file = match.group(1).strip().lower()
    if 'testlib' in file or 'jngen' in file or 'stresslib' in file:
        return False
    if file.endswith('.h'):
        return False
    return True


def _check_for_compilation_warnings(
    sandbox: SandboxBase, stderr_file: Optional[pathlib.Path]
) -> bool:
    if stderr_file is None:
        return False
    if not sandbox.file_exists(stderr_file):
        return False
    with sandbox.get_file(stderr_file) as f:
        return any(
            _check_for_compilation_warnings_in_line(line.strip().decode()) for line in f
        )


def _build_program_error_run_log(
    e: ProgramError, metadata: Optional[RunLogMetadata] = None
) -> RunLog:
    return RunLog(
        exitcode=1,
        exitstatus=SandboxBase.EXIT_SANDBOX_ERROR,
        time=0.0,
        memory=0,
        metadata=metadata,
        sandbox=str(e),
        exitindex=0,
    )


def _build_run_log(
    sandbox_log: SandboxLog,
    sandbox: SandboxBase,
    params: SandboxParams,
    metadata: Optional[RunLogMetadata] = None,
) -> RunLog:
    execution_time = sandbox_log.execution_time
    if execution_time is not None and (
        sandbox_log.exitstatus == SandboxBase.EXIT_TIMEOUT
        or sandbox_log.exitstatus == SandboxBase.EXIT_TIMEOUT_WALL
    ):
        execution_time = max(execution_time, (params.timeout or 0.0) / 1000)

    run_log = RunLog(
        exitcode=sandbox_log.exitcode,
        exitstatus=sandbox_log.exitstatus,
        time=execution_time,
        memory=sandbox_log.memory_used,
        metadata=metadata,
        sandbox=sandbox_log.dump_other_logs(),
        exitindex=sandbox_log.exit_index,
    )
    if metadata is not None and metadata.is_sanitized:
        run_log.warnings = _check_for_sanitizer_warnings(
            sandbox,
            params.stderr_file,
        )
    return run_log


class CompilationError(RbxException):
    pass


def compile(
    commands: List[str],
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
):
    sandbox.reset()

    commands = _try_following_alias_for_commands(commands)
    _process_input_artifacts(artifacts, sandbox)

    if not commands:
        # Code does not need preprocessing of any kind.
        return

    logs: List[PreprocessLog] = []
    params = params.model_copy(deep=True)  # Copy to allow further modification.

    for i, command in enumerate(commands):
        _maybe_complain_about_sanitization(command)
        cmd = _split_and_expand(command, sandbox, params)
        stdout_file = pathlib.PosixPath(f'compile-{i}.stdout')
        stderr_file = pathlib.PosixPath(f'compile-{i}.stderr')
        params.set_stdall(stdout=stdout_file, stderr=stderr_file)

        # Remove memory constraints for Java.
        if is_java_like_command(get_exe_from_command(command)):
            params.address_space = None

        try:
            sandbox_log = sandbox.run(cmd, params)
        except ProgramError as e:
            with CompilationError() as err:
                err.print(
                    '[error]FAILED[/error] Preprocessing failed with command',
                    utils.highlight_json_obj(cmd),
                )
                err.print(e)

        std_outputs = [
            sandbox.get_file_to_string(stderr_file, maxlen=None)
            if sandbox.file_exists(stderr_file)
            else '<No stderr produced by command>',
            sandbox.get_file_to_string(stdout_file, maxlen=None)
            if sandbox.file_exists(stdout_file)
            else '<No stdout produced by command>',
        ]

        log = PreprocessLog(
            cmd=cmd,
            exitcode=sandbox_log.exitcode,
            exitstatus=sandbox_log.exitstatus,
            time=sandbox_log.execution_time,
            memory=sandbox_log.memory_used,
            warnings=_check_for_compilation_warnings(sandbox, stderr_file),
            log='\n'.join(std_outputs),
            sandbox=sandbox_log.dump_other_logs(),
        )
        logs.append(log)

        if log.exitcode != 0:
            break

    if artifacts.logs is not None:
        artifacts.logs.preprocess = logs

    if logs and logs[-1].exitcode != 0:
        with CompilationError() as err:
            err.print(
                '[error]FAILED[/error] Preprocessing failed with command',
                utils.highlight_json_obj(logs[-1].cmd),
            )
            err.print(f'[error]Summary:[/error] {logs[-1].get_summary()}')
            err.print(Text.from_ansi(logs[-1].log), style='default')

    # testing_utils.print_directory_tree(sandbox.get_root_path())
    if not _process_output_artifacts(artifacts, sandbox):
        raise CompilationError()


async def run(
    command: str,
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    metadata: Optional[RunLogMetadata] = None,
) -> Optional[RunLog]:
    sandbox.reset()

    _process_input_artifacts(artifacts, sandbox)
    _process_fifos(artifacts, sandbox)
    cmd = _split_and_expand(command, sandbox, params)
    params = params.model_copy(deep=True)  # Copy to allow further modification.

    # Remove memory constraints for Java.
    if is_java_like_command(get_exe_from_command(command)):
        params.address_space = None

    try:
        sandbox_log = await asyncio.to_thread(sandbox.run, cmd, params)

        if not _process_output_artifacts(artifacts, sandbox):
            return None

        run_log = _build_run_log(sandbox_log, sandbox, params, metadata)
    except ProgramError as e:
        run_log = _build_program_error_run_log(e, metadata)

    if artifacts.logs is not None:
        artifacts.logs.run = run_log.model_copy()
    return run_log


@dataclasses.dataclass
class CoordinatedRunParams:
    command: str
    params: SandboxParams
    metadata: Optional[RunLogMetadata] = None


async def run_coordinated(
    interactor: CoordinatedRunParams,
    solution: CoordinatedRunParams,
    artifacts: GradingArtifacts,
    sandbox: SandboxBase,
    merged_capture: Optional[pathlib.Path] = None,
    line_capture: bool = False,
) -> Tuple[Optional[RunLog], Optional[RunLog]]:
    sandbox.reset()

    _process_input_artifacts(artifacts, sandbox)
    _process_fifos(artifacts, sandbox)

    interactor_cmd = _split_and_expand(interactor.command, sandbox, interactor.params)
    solution_cmd = _split_and_expand(solution.command, sandbox, solution.params)

    interactor_params = interactor.params.model_copy(deep=True)
    solution_params = solution.params.model_copy(deep=True)

    if is_java_like_command(get_exe_from_command(solution.command)):
        solution_params.address_space = None

    try:
        solution_sandbox_log, interactor_sandbox_log = sandbox.run_communication(
            solution_cmd,
            solution_params,
            interactor_cmd,
            interactor_params,
            CommunicationParams(
                merged_capture=merged_capture,
                tee_mode='line' if line_capture else 'char',
            ),
        )

        if not _process_output_artifacts(artifacts, sandbox):
            return None, None

        solution_log = _build_run_log(
            solution_sandbox_log, sandbox, solution.params, solution.metadata
        )
        interactor_log = _build_run_log(
            interactor_sandbox_log, sandbox, interactor.params, interactor.metadata
        )
    except ProgramError as e:
        solution_log = _build_program_error_run_log(e, solution.metadata)
        interactor_log = _build_program_error_run_log(e, interactor.metadata)

    if artifacts.logs is not None:
        artifacts.logs.run = solution_log
        artifacts.logs.interactor_run = interactor_log

    return solution_log, interactor_log


def get_checker_sandbox_params() -> SandboxParams:
    params = SandboxParams(
        max_processes=None,
        preserve_env=True,
    )
    params.add_mapped_directory(pathlib.Path('/usr'))
    params.add_mapped_directory(pathlib.Path('/etc'))
    return params



import pathlib
from typing import Dict, Iterator, Tuple
from unittest import mock

import pytest
import typer

from rbx.box import checkers
from rbx.box.checkers import compile_checker, compile_interactor
from rbx.box.schema import Checker, CodeItem, Testcase
from rbx.box.testing import testing_package
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.limits import Limits
from rbx.grading.steps import (
    CompilationError,
    DigestOrSource,
    Outcome,
    RunLog,
    RunLogMetadata,
)
from rbx.utils import StatusProgress

INTERESTING_CHECKERS = [
    'checkers/checker.cpp',
    'checkers/checker-fl.cpp',
    'checkers/checker-crash.cpp',
]


@pytest.fixture(scope='module')
def pkg_with_compiled_checker(tmp_path_factory):
    pkg_dir = tmp_path_factory.mktemp('pkg')
    with testing_package.TestingPackage(pkg_dir.absolute()) as testing_pkg:
        checkers = {}
        for checker in INTERESTING_CHECKERS:
            testing_pkg.set_checker('checker.cpp', src=checker)
            checkers[checker] = compile_checker()
        yield testing_pkg, checkers


@pytest.fixture(autouse=True)
def checker_pkg(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, str],
    testing_pkg: testing_package.TestingPackage,
) -> Iterator[testing_package.TestingPackage]:
    pkg, _ = pkg_with_compiled_checker
    testing_pkg.copy_from(pkg)
    with testing_pkg:
        yield testing_pkg


@pytest.fixture
def checker_digest(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, Dict[str, str]],
) -> str:
    _, checkers = pkg_with_compiled_checker
    return checkers['checkers/checker.cpp']


@pytest.fixture
def checker_digest_dict(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, Dict[str, str]],
) -> Dict[str, str]:
    _, checkers = pkg_with_compiled_checker
    return checkers


@pytest.fixture
def testcase(tmp_path_factory) -> Testcase:
    testcase_dir = tmp_path_factory.mktemp('testcase')
    input_path = testcase_dir / 'input.txt'
    input_path.touch()
    output_path = testcase_dir / 'output.txt'
    output_path.touch()
    return Testcase(
        inputPath=input_path,
        outputPath=output_path,
    )


@pytest.fixture
def testcase_no_output(tmp_path_factory) -> Testcase:
    testcase_dir = tmp_path_factory.mktemp('testcase')
    input_path = testcase_dir / 'input.txt'
    input_path.touch()
    return Testcase(
        inputPath=input_path,
        outputPath=None,
    )


@pytest.fixture
def program_output(tmp_path_factory) -> pathlib.Path:
    output_dir = tmp_path_factory.mktemp('output')
    output_path = output_dir / 'output.txt'
    output_path.touch()
    return output_path


@pytest.fixture
def run_log() -> RunLog:
    return RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_OK,
        metadata=RunLogMetadata(
            limits=Limits(time=1000, memory=1024, output=1024, isDoubleTL=True),
            timeLimit=1000,
            memoryLimit=1024,
            retryIndex=0,
        ),
    )


@pytest.fixture
def run_log_with_warnings(run_log: RunLog) -> RunLog:
    run_log.warnings = True
    return run_log


# Compilation tests
@mock.patch('rbx.box.code.compile_item')
@mock.patch('rbx.box.package.get_checker')
def test_compile_checker_success(
    mock_get_checker: mock.Mock,
    mock_compile_item: mock.Mock,
) -> None:
    mock_get_checker.return_value = CodeItem(path=pathlib.Path('checker.cpp'))
    mock_compile_item.return_value = 'test_digest'

    result = compile_checker()

    assert result == 'test_digest'
    mock_get_checker.assert_called_once()
    mock_compile_item.assert_called_once()


@mock.patch('rbx.box.code.compile_item')
@mock.patch('rbx.box.package.get_checker')
def test_compile_checker_failure(
    mock_get_checker: mock.Mock,
    mock_compile_item: mock.Mock,
) -> None:
    mock_get_checker.return_value = CodeItem(path=pathlib.Path('checker.cpp'))
    mock_compile_item.side_effect = CompilationError()

    with pytest.raises(CompilationError):
        compile_checker()


@mock.patch('rbx.box.code.compile_item')
@mock.patch('rbx.box.package.get_interactor')
def test_compile_interactor_success(
    mock_get_interactor: mock.Mock,
    mock_compile_item: mock.Mock,
) -> None:
    mock_get_interactor.return_value = CodeItem(path=pathlib.Path('interactor.cpp'))
    mock_compile_item.return_value = 'test_digest'

    result = compile_interactor()

    assert result == 'test_digest'
    mock_get_interactor.assert_called_once()
    mock_compile_item.assert_called_once()


@mock.patch('rbx.box.code.compile_item')
@mock.patch('rbx.box.package.get_interactor')
def test_compile_interactor_with_progress(
    mock_get_interactor: mock.Mock,
    mock_compile_item: mock.Mock,
) -> None:
    mock_get_interactor.return_value = CodeItem(path=pathlib.Path('interactor.cpp'))
    mock_compile_item.return_value = 'test_digest'
    progress = mock.Mock(spec=StatusProgress)

    result = compile_interactor(progress)

    assert result == 'test_digest'
    progress.update.assert_called_once_with('Compiling interactor...')


@mock.patch('rbx.box.package.get_interactor')
def test_compile_interactor_not_found(
    mock_get_interactor: mock.Mock,
) -> None:
    mock_get_interactor.return_value = None

    with pytest.raises(typer.Exit):
        compile_interactor()


@mock.patch('rbx.box.code.compile_item')
@mock.patch('rbx.box.package.get_interactor')
def test_compile_interactor_failure(
    mock_get_interactor: mock.Mock,
    mock_compile_item: mock.Mock,
) -> None:
    mock_get_interactor.return_value = CodeItem(path=pathlib.Path('interactor.cpp'))
    mock_compile_item.side_effect = Exception('Compilation failed')

    with pytest.raises(typer.Exit):
        compile_interactor()


# Test check_with_no_output function
def test_check_with_no_output_none() -> None:
    result = checkers.check_with_no_output(None)
    assert result.outcome == Outcome.INTERNAL_ERROR


def test_check_with_no_output_accepted(run_log: RunLog) -> None:
    result = checkers.check_with_no_output(run_log)
    assert result.outcome == Outcome.ACCEPTED


def test_check_with_no_output_tle(run_log: RunLog) -> None:
    run_log.exitstatus = SandboxBase.EXIT_TIMEOUT
    result = checkers.check_with_no_output(run_log)
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED


def test_check_with_no_output_soft_tle(run_log: RunLog) -> None:
    run_log.time = 1.5  # Greater than TL but less than 2*TL
    result = checkers.check_with_no_output(run_log)
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    assert result.no_tle_outcome == Outcome.ACCEPTED


# Test with testcase that has no output file
async def test_check_with_testcase_no_output(
    checker_digest: str,
    testcase_no_output: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    # When there's no expected output, program output should also be empty
    program_output.write_text('')
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase_no_output,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED


# Test sanitizer warnings
async def test_check_sets_sanitizer_warnings(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log_with_warnings: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    result = await checkers.check(
        checker_digest,
        run_log_with_warnings,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED
    assert result.sanitizer_warnings is True


async def test_check_no_sanitizer_warnings(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED
    assert result.sanitizer_warnings is False


# Test output size calculation edge cases
async def test_program_output_exactly_at_limit(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    # Write exactly 1024 KB (1024 * 1024 bytes)
    content = 'a' * (1024 * 1024)
    testcase.outputPath.write_text(content)
    program_output.write_text(content)

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_program_output_just_over_limit(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    # Write just over 1024 KB (1024 * 1024 + 1 bytes)
    program_output.write_text('a' * (1024 * 1024 + 1))

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.OUTPUT_LIMIT_EXCEEDED


# Test checker with different exit status combinations
@mock.patch('rbx.box.code.run_item')
async def test_checker_with_timeout_exit_status(
    mock_run_item: mock.AsyncMock,
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    mock_run_item.return_value = RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_TIMEOUT,
    )

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.JUDGE_FAILED
    assert 'checker failed with exit status' in result.message


@mock.patch('rbx.box.code.run_item')
async def test_checker_with_signal_exit_status(
    mock_run_item: mock.AsyncMock,
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    mock_run_item.return_value = RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_SIGNAL,
    )

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.JUDGE_FAILED
    assert 'checker failed with exit status' in result.message


async def test_check_fails_with_no_run_log(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')
    result = await checkers.check(
        checker_digest,
        None,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.INTERNAL_ERROR


async def test_wcmp_checker_works_with_skipped_no_run_log(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')
    result = await checkers.check(
        checker_digest,
        None,
        testcase,
        program_output,
        skip_run_log=True,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_wcmp_checker_works_with_skipped_bad_run_log(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
        skip_run_log=True,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_wcmp_checker_works(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_wcmp_checker_wa(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('456\n')
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.WRONG_ANSWER


async def test_run_log_has_sandbox_tle(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_TIMEOUT
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED


async def test_run_log_has_sandbox_idleness(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_TIMEOUT_WALL
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.IDLENESS_LIMIT_EXCEEDED


async def test_run_log_has_sandbox_mle(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.MEMORY_LIMIT_EXCEEDED


async def test_run_log_has_sandbox_rte(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.RUNTIME_ERROR


async def test_run_log_has_sandbox_signal(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_SIGNAL
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.RUNTIME_ERROR


async def test_run_log_has_sandbox_output_limit(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_OUTPUT_LIMIT_EXCEEDED
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.OUTPUT_LIMIT_EXCEEDED


async def test_run_log_has_sandbox_internal_error(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.exitstatus = SandboxBase.EXIT_SANDBOX_ERROR
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.INTERNAL_ERROR


async def test_run_log_has_time_based_2tle(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.time = 2.5  # Greater than 2*TL = 2s
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED


async def test_run_log_has_unbounded_tl(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.time = 2.5  # Greater than 2*TL = 2s
    assert run_log.metadata is not None
    run_log.metadata.timeLimit = None  # TL is unbounded
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_run_log_has_unbounded_tl_because_sanitized(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.time = 2.5  # Greater than 2*TL = 2s
    assert run_log.metadata is not None
    run_log.metadata.is_sanitized = True
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.ACCEPTED


async def test_run_log_has_soft_tle(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.time = 1.5  # Less than 2*TL = 2s
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    assert result.no_tle_outcome == Outcome.ACCEPTED


async def test_run_log_has_soft_tle_but_wa(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('456\n')

    run_log.time = 1.5  # Less than 2*TL = 2s
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    assert result.no_tle_outcome == Outcome.WRONG_ANSWER


async def test_run_log_has_soft_tle_but_rte(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    run_log.time = 1.5  # Less than 2*TL = 2s
    run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN
    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    assert result.no_tle_outcome == Outcome.RUNTIME_ERROR


async def test_program_output_is_too_large(
    checker_digest: str,
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    assert testcase.outputPath
    program_output.write_text('123\n' * 10**6)

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.OUTPUT_LIMIT_EXCEEDED


async def test_checker_item_run_log_has_fl_exitcode(
    checker_digest_dict: Dict[str, str],
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    checker_digest = checker_digest_dict['checkers/checker-fl.cpp']
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.JUDGE_FAILED


async def test_checker_item_run_log_has_invalid_exitcode(
    checker_digest_dict: Dict[str, str],
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    checker_digest = checker_digest_dict['checkers/checker-crash.cpp']
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    result = await checkers.check(
        checker_digest,
        run_log,
        testcase,
        program_output,
    )
    assert result.outcome == Outcome.JUDGE_FAILED
    assert 'checker failed with unknown exit code 42' in result.message


@mock.patch('rbx.box.code.run_item')
async def test_checker_item_run_log_is_none(
    mock_run_item: mock.AsyncMock,
    checker_digest_dict: Dict[str, str],
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    checker_digest = checker_digest_dict['checkers/checker.cpp']
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    mock_run_item.return_value = None

    with pytest.raises(typer.Exit):  # noqa: F821
        await checkers.check(
            checker_digest,
            run_log,
            testcase,
            program_output,
        )

    mock_run_item.assert_awaited_with(
        Checker(
            path=pathlib.Path('checker.cpp'),
        ),
        DigestOrSource.create(checker_digest),
        stderr=mock.ANY,
        inputs=mock.ANY,
        extra_args=mock.ANY,
    )


@mock.patch('rbx.box.code.run_item')
async def test_checker_item_has_sandbox_error(
    mock_run_item: mock.AsyncMock,
    checker_digest_dict: Dict[str, str],
    testcase: Testcase,
    program_output: pathlib.Path,
    run_log: RunLog,
) -> None:
    checker_digest = checker_digest_dict['checkers/checker.cpp']
    assert testcase.outputPath
    testcase.outputPath.write_text('123\n')
    program_output.write_text('123\n')

    mock_run_item.return_value = RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_SANDBOX_ERROR,
    )

    with pytest.raises(typer.Exit):  # noqa: F821
        await checkers.check(
            checker_digest,
            run_log,
            testcase,
            program_output,
        )

    mock_run_item.assert_awaited_with(
        Checker(
            path=pathlib.Path('checker.cpp'),
        ),
        DigestOrSource.create(checker_digest),
        stderr=mock.ANY,
        inputs=mock.ANY,
        extra_args=mock.ANY,
    )


# Test that process_checker_run_log uses last line of message
def test_process_checker_run_log_uses_last_line() -> None:
    """Test that process_checker_run_log extracts the last line from multi-line messages."""
    checker_run_log = RunLog(
        exitcode=1,
        exitstatus=SandboxBase.EXIT_NONZERO_RETURN,
    )
    multi_line_message = 'Debug info\nMore debug\nActual error message'

    result = checkers.process_checker_run_log(checker_run_log, multi_line_message)

    assert result.outcome == Outcome.WRONG_ANSWER
    assert result.message == 'Actual error message'



from __future__ import annotations

from dataclasses import dataclass
from typing import List, NamedTuple, Optional

from textual import events, on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.message import Message
from textual.widgets import Label, ListItem, ListView


class MenuItem(NamedTuple):
    """An entry in a Menu."""

    description: str
    action: str
    key: Optional[str] = None


class _MenuOptionLabel(Label):
    ALLOW_SELECT = False


class MenuOption(ListItem):
    ALLOW_SELECT = False

    def __init__(self, description: str, key: Optional[str]) -> None:
        self._description = description
        self._key = key
        super().__init__()

    def compose(self) -> ComposeResult:
        yield _MenuOptionLabel(self._key or ' ', id='key')
        yield _MenuOptionLabel(self._description, id='description')


class Menu(ListView, can_focus=True):
    BINDINGS = [Binding('escape', 'dismiss', 'Dismiss')]

    DEFAULT_CSS = """
    Menu {
        width: auto;
        height: auto;
        max-width: 100%;
        overlay: screen;
        position: absolute;
        color: $foreground;
        background: $panel;
        border: round $accent;
        constrain: inside inside;
        padding: 0;

        & > MenuOption {
            layout: horizontal;
            width: 1fr;
            padding: 0 1;
            height: auto !important;
            overflow: auto;
            #description {
                color: $text 80%;
                width: 1fr;
            }
            #key {
                padding-right: 1;
                text-style: bold;
            }
        }

        &:blur {
            background-tint: transparent;
            & > ListItem.-highlight {
                color: $block-cursor-blurred-foreground;
                background: $block-cursor-blurred-background 30%;
                text-style: $block-cursor-blurred-text-style;
            }
        }

        &:focus {
            background-tint: transparent;
            & > ListItem.-highlight {
                color: $block-cursor-blurred-foreground;
                background: $block-cursor-blurred-background;
                text-style: $block-cursor-blurred-text-style;
            }
        }
    }
    """

    @dataclass
    class OptionSelected(Message):
        menu: Menu
        action: str

    @dataclass
    class Dismissed(Message):
        menu: Menu

    def __init__(self, options: List[MenuItem], *args, **kwargs) -> None:
        self._options = options
        self._dismissed = False
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        self.extend(MenuOption(item.description, item.key) for item in self._options)

    async def _activate_index(self, index: int) -> None:
        if self._dismissed or index < 0 or index >= len(self._options):
            return
        action = self._options[index].action
        self.post_message(self.OptionSelected(self, action))

    def _dismiss(self) -> None:
        if self._dismissed:
            return
        self._dismissed = True
        self.post_message(self.Dismissed(self))

    async def action_dismiss(self) -> None:
        self._dismiss()

    async def on_blur(self) -> None:
        self._dismiss()

    @on(events.Key)
    async def _on_key(self, event: events.Key) -> None:
        for index, option in enumerate(self._options):
            if option.key is not None and event.key == option.key:
                self.index = index
                event.stop()
                await self._activate_index(index)
                break

    @on(ListView.Selected)
    async def _on_selected(self, event: ListView.Selected) -> None:
        event.stop()
        if event.index is not None:
            await self._activate_index(event.index)



import dataclasses
import enum
import shlex
from time import monotonic
from typing import List, Optional, Tuple

from textual import events, on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.message import Message
from textual.screen import ModalScreen
from textual.selection import Selection
from textual.widget import Widget
from textual.widgets import Footer, Header, Input, Label, ListItem, ListView, Select

from rbx.box.ui._vendor.toad.widgets.command_pane import CommandPane
from rbx.box.ui.main import rbxBaseApp
from rbx.box.ui.screens.tab_selector import TabSelectorModal
from rbx.box.ui.task_queue import Task, TaskQueue
from rbx.box.ui.widgets.menu import Menu, MenuItem

_ESCAPE_TAP_DURATION = 0.4


class _AppCommandPane(CommandPane):
    """CommandPane that redirects focus to sidebar on blur (double-escape)."""

    def blur(self):
        try:
            sidebar = self.screen.query_one('#command-list', ListView)
            self.screen.set_focus(sidebar)
        except Exception:
            super().blur()
        return self

    def on_blur(self) -> None:
        self.border_subtitle = '[b]tab[/b] to focus'

    def selection_updated(self, selection: Selection | None) -> None:
        super().selection_updated(selection)
        if self.has_focus and selection is not None:
            self.border_subtitle = '[b]ctrl+y[/b] copy selection'
        elif self.has_focus:
            self.border_subtitle = 'Tap [b]esc[/b] [i]twice[/i] to exit'

    async def on_key(self, event: events.Key) -> None:
        if event.key == 'ctrl+y':
            selected = self.screen.get_selected_text()
            if selected:
                self.app.copy_to_clipboard(selected)
                self.screen.clear_selection()
                self.border_subtitle = 'Tap [b]esc[/b] [i]twice[/i] to exit'
                event.stop()
                event.prevent_default()
                return
        await super().on_key(event)


class ShellInput(Input):
    """Input that captures Tab/Shift+Tab and supports double-Escape to exit."""

    class Escaped(Message):
        """Posted when the user double-taps Escape to exit the input."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._escaping = False
        self._escape_time: float = 0.0
        self._escape_timer = None

    def _reset_escaping(self) -> None:
        self._escaping = False

    def on_focus(self) -> None:
        self.border_subtitle = _INPUT_FOCUSED_SUBTITLE

    def on_blur(self) -> None:
        self.border_subtitle = _INPUT_BLURRED_SUBTITLE

    def _cancel_escape_timer(self) -> None:
        if self._escape_timer is not None:
            self._escape_timer.stop()
            self._escape_timer = None

    def on_key(self, event: events.Key) -> None:
        if event.key in ('tab', 'shift+tab'):
            event.stop()
            event.prevent_default()
            return
        if event.key == 'escape':
            event.stop()
            event.prevent_default()
            self._cancel_escape_timer()
            if (
                self._escaping
                and monotonic() < self._escape_time + _ESCAPE_TAP_DURATION
            ):
                self._escaping = False
                self.post_message(self.Escaped())
            else:
                self._escaping = True
                self._escape_time = monotonic()
                self._escape_timer = self.set_timer(
                    _ESCAPE_TAP_DURATION, self._reset_escaping
                )


_SIDEBAR_SUBTITLE = '[b]?[/b] help'
_INPUT_FOCUSED_SUBTITLE = '[b]enter[/b] run  [b]esc\u00d72[/b] cancel'
_INPUT_BLURRED_SUBTITLE = '[b]![/b] to focus'
_SELECT_SUBTITLE = '[b]\u25c2\u25b8[/b] sub-cmd'


class HelpModal(ModalScreen[None]):
    BINDINGS = [
        ('escape', 'app.pop_screen', 'Close'),
        ('question_mark', 'app.pop_screen', 'Close'),
    ]

    DEFAULT_CSS = """
    HelpModal {
        align: center middle;
    }
    #help-dialog {
        max-width: 60;
        height: auto;
        padding: 1 2;
        border: solid $accent;
        background: $surface;
    }
    #help-dialog Label {
        width: 1fr;
        margin-bottom: 1;
    }
    #help-title {
        text-style: bold;
        text-align: center;
    }
    #help-hints {
        text-align: center;
        color: $text 60%;
    }
    """

    def compose(self) -> ComposeResult:
        with Container(id='help-dialog'):
            yield Label('Keyboard Shortcuts', id='help-title')
            yield Label(
                '[b]Sidebar (command list)[/b]\n'
                '  [b]tab[/b]         Focus terminal\n'
                '  [b]![/b]           Open shell input\n'
                '  [b]\u2190 / \u2192[/b]       Previous / next sub-command\n'
                '  [b]?[/b]           Show this help\n'
                '  [b]q[/b]           Quit',
                markup=True,
            )
            yield Label(
                '[b]Terminal[/b]\n'
                '  [b]esc\u00d72[/b]        Return to sidebar\n'
                '  [b]ctrl+y[/b]      Copy selected text\n'
                '  (all other keys go to the running process)',
                markup=True,
            )
            yield Label(
                '[b]Shell input[/b]\n'
                '  [b]enter[/b]        Submit command\n'
                '  [b]esc\u00d72[/b]        Cancel and return to sidebar',
                markup=True,
            )
            yield Label(
                '[b]esc[/b] or [b]?[/b] to close',
                id='help-hints',
            )


class CommandStatus(enum.Enum):
    PENDING = 'pending'
    RUNNING = 'running'
    SUCCESS = 'success'
    FAILED = 'failed'


_STATUS_MARKUP = {
    CommandStatus.PENDING: '[dim]○[/dim]',
    CommandStatus.RUNNING: '[yellow]●[/yellow]',
    CommandStatus.SUCCESS: '[green]✓[/green]',
    CommandStatus.FAILED: '[red]✗[/red]',
}

_STATUS_ICON = {
    CommandStatus.PENDING: '○',
    CommandStatus.RUNNING: '●',
    CommandStatus.SUCCESS: '✓',
    CommandStatus.FAILED: '✗',
}


@dataclasses.dataclass
class CommandEntry:
    argv: List[str]
    name: Optional[str] = None
    cwd: Optional[str] = None
    prefix: Optional[str] = None
    placeholder_prefix: Optional[str] = None

    @property
    def display_name(self) -> str:
        return self.name if self.name else ' '.join(self.argv)

    def make_raw_shell_command(self, cmd: str) -> str:
        if self.prefix is not None:
            cmd = f'{self.prefix} {cmd}'
        if self.cwd is not None:
            cmd = f'cd {shlex.quote(self.cwd)} && exec {cmd}'
        return cmd

    def make_shell_command(self, argv: List[str]) -> str:
        return self.make_raw_shell_command(shlex.join(argv))

    @property
    def shell_command(self) -> str:
        return self.make_shell_command(self.argv)


@dataclasses.dataclass
class SubCommand:
    name: str
    shell_command: str
    pane_id: str
    status: CommandStatus = CommandStatus.PENDING
    task_id: Optional[int] = None


class TabState:
    def __init__(self, entry: CommandEntry, tab_index: int):
        self.entry = entry
        self.tab_index = tab_index
        self.sub_commands: List[SubCommand] = []
        self._next_sub_id = 0

    def _append_sub_command(self, name: str, shell_command: str) -> SubCommand:
        pane_id = f'cmd-pane-{self.tab_index}-{self._next_sub_id}'
        sub = SubCommand(
            name=name,
            shell_command=shell_command,
            pane_id=pane_id,
        )
        self._next_sub_id += 1
        self.sub_commands.append(sub)
        return sub

    def add_sub_command(self, name: str, argv: List[str]) -> SubCommand:
        shell_command = self.entry.make_shell_command(argv)
        return self._append_sub_command(name, shell_command)

    def add_sub_command_raw(self, name: str, raw_command: str) -> SubCommand:
        shell_command = self.entry.make_raw_shell_command(raw_command)
        return self._append_sub_command(name, shell_command)

    @property
    def is_idle(self) -> bool:
        if not self.sub_commands:
            return True
        return all(
            s.status in (CommandStatus.SUCCESS, CommandStatus.FAILED)
            for s in self.sub_commands
        )

    @property
    def aggregate_status(self) -> CommandStatus:
        if not self.sub_commands:
            return CommandStatus.PENDING
        statuses = {s.status for s in self.sub_commands}
        for status in (
            CommandStatus.FAILED,
            CommandStatus.RUNNING,
            CommandStatus.PENDING,
        ):
            if status in statuses:
                return status
        return CommandStatus.SUCCESS


class rbxCommandApp(rbxBaseApp):
    class TaskReady(Message):
        def __init__(self, task: Task):
            self.task = task
            super().__init__()

    TITLE = 'rbx'
    CSS_PATH = 'css/app.tcss'
    DEFAULT_CSS = """
    #command-app {
        height: 1fr;
    }
    #command-list-container {
        min-width: 20;
        max-width: 40;
        height: 1fr;
    }
    #command-list {
        width: 1fr;
    }
    #command-list:focus {
        border: solid dodgerblue;
    }
    #command-display-area {
        height: 1fr;
        width: 1fr;
    }
    #command-select {
        width: 1fr;
        margin: 0;
    }
    #command-pane-container {
        height: 1fr;
        width: 1fr;
    }
    #command-pane-container CommandPane {
        height: 1fr;
        border: solid $accent;
        padding: 0 1;
        scrollbar-size-vertical: 1;
    }
    #command-pane-container CommandPane:focus {
        border: solid dodgerblue;
    }
    #command-input-container {
        dock: bottom;
        height: auto;
    }
    #command-input-prefix {
        width: auto;
        height: 3;
        content-align: left middle;
        padding: 0 1;
        color: $accent;
    }
    #command-input {
        width: 1fr;
    }
    #command-input:focus {
        border: tall dodgerblue;
    }
    """
    BINDINGS = [
        ('q', 'quit', 'Quit'),
    ]

    def __init__(self, commands: List[CommandEntry], parallel: bool = False):
        super().__init__()
        self.commands = commands
        self.parallel = parallel
        self._tabs: List[TabState] = []
        self._active_tab: int = 0
        self._task_queue = TaskQueue(
            num_terminals=len(commands),
            parallel=parallel,
            on_task_ready=lambda t: self.post_message(self.TaskReady(t)),
        )
        self._pending_command: Optional[str] = None

        # Initialize tab states and add initial sub-commands.
        for i, cmd in enumerate(commands):
            tab = TabState(entry=cmd, tab_index=i)
            if cmd.argv:
                tab.add_sub_command(name=' '.join(cmd.argv), argv=cmd.argv)
            self._tabs.append(tab)

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Horizontal(id='command-app'):
            with Vertical(id='command-list-container'):
                yield ListView(
                    *[
                        ListItem(
                            Label(self._make_tab_label(i), markup=True),
                            id=f'cmd-item-{i}',
                        )
                        for i in range(len(self.commands))
                    ],
                    id='command-list',
                )
            with Vertical(id='command-display-area'):
                yield Select[int](
                    self._get_select_options(0) or [('—', -1)],
                    id='command-select',
                    allow_blank=False,
                )
                yield Vertical(id='command-pane-container')
                with Horizontal(id='command-input-container'):
                    yield Label(
                        self._get_input_prefix_text(0),
                        id='command-input-prefix',
                    )
                    yield ShellInput(
                        id='command-input',
                        placeholder=self._get_input_placeholder(0),
                    )

    def _make_tab_label(self, index: int) -> str:
        tab = self._tabs[index]
        icon = _STATUS_MARKUP[tab.aggregate_status]
        name = tab.entry.display_name
        return f'{icon} {name}'

    def _update_sidebar(self, index: int):
        item = self.query_one(f'#cmd-item-{index}', ListItem)
        label = item.query_one(Label)
        label.update(self._make_tab_label(index))

    def _get_select_options(self, tab_index: int) -> List[Tuple[str, int]]:
        return [
            (f'{_STATUS_ICON[sub.status]} {sub.name}', i)
            for i, sub in enumerate(self._tabs[tab_index].sub_commands)
        ]

    def _get_input_prefix_text(self, tab_index: int) -> str:
        return self._tabs[tab_index].entry.prefix or ''

    def _update_input_prefix(self, tab_index: int):
        prefix_label = self.query_one('#command-input-prefix', Label)
        tab = self._tabs[tab_index]
        if tab.entry.prefix is not None:
            prefix_label.update(tab.entry.prefix)
            prefix_label.display = True
        else:
            prefix_label.update('')
            prefix_label.display = False

    def _get_input_placeholder(self, tab_index: int) -> str:
        tab = self._tabs[tab_index]
        if tab.entry.placeholder_prefix is not None:
            return f'{tab.entry.placeholder_prefix} <command>'
        return 'Type a command and press Enter...'

    def _refresh_select(self):
        select = self.query_one('#command-select', Select)
        options = self._get_select_options(self._active_tab)
        current_value = select.value
        select.set_options(options or [('—', -1)])
        # Try to preserve selection; if it no longer exists, select last.
        if any(v == current_value for _, v in options):
            select.value = current_value
        elif options:
            select.value = options[-1][1]

    def _show_pane(self, pane_id: str):
        container = self.query_one('#command-pane-container', Vertical)
        for child in container.query(CommandPane):
            child.display = child.id == pane_id

    def _get_selected_pane_id(self) -> Optional[str]:
        select = self.query_one('#command-select', Select)
        if select.value is Select.BLANK:
            return None
        sub_index: int = select.value  # type: ignore[assignment]
        tab = self._tabs[self._active_tab]
        if 0 <= sub_index < len(tab.sub_commands):
            return tab.sub_commands[sub_index].pane_id
        return None

    def on_mount(self):
        sidebar = self.query_one('#command-list', ListView)
        sidebar.border_title = 'Commands'
        sidebar.border_subtitle = _SIDEBAR_SUBTITLE

        select = self.query_one('#command-select', Select)
        select.border_subtitle = _SELECT_SUBTITLE

        # Mount initial CommandPanes.
        container = self.query_one('#command-pane-container', Vertical)
        for tab in self._tabs:
            for sub in tab.sub_commands:
                pane = _AppCommandPane(id=sub.pane_id)
                pane.border_title = sub.name
                container.mount(pane)

        # Show first tab's first pane.
        self._switch_tab(0)

        self.watch(
            self.query_one('#command-list', ListView),
            'index',
            self._on_tab_selected,
        )

        # Redirect to sidebar when focus becomes None (e.g. modal dismiss).
        self.watch(self.screen, 'focused', self._on_focused_changed)

        # Initial focus on the sidebar.
        self._focus_sidebar()

        # Enqueue initial commands.
        for i, tab in enumerate(self._tabs):
            for sub in tab.sub_commands:
                task = self._task_queue.enqueue(sub.shell_command, terminal_id=i)
                sub.task_id = task.task_id

    def on_rbx_command_app_task_ready(self, event: TaskReady) -> None:
        task = event.task
        tab = self._tabs[task.terminal_id]
        # Find the sub-command linked to this task.
        sub = next((s for s in tab.sub_commands if s.task_id == task.task_id), None)
        if sub is None:
            return
        sub.status = CommandStatus.RUNNING
        self._update_sidebar(task.terminal_id)
        self._refresh_select_if_active(task.terminal_id)
        pane = self.query_one(f'#{sub.pane_id}', CommandPane)
        pane.execute(task.command)

    def _on_tab_selected(self, index: Optional[int]):
        if index is None:
            return
        self._switch_tab(index)

    def _switch_tab(self, index: int):
        self._active_tab = index
        self._refresh_select()

        # Select the latest sub-command by default.
        tab = self._tabs[index]
        select = self.query_one('#command-select', Select)
        if tab.sub_commands:
            select.value = len(tab.sub_commands) - 1
            self._show_pane(tab.sub_commands[-1].pane_id)
        else:
            # Hide all panes for this tab.
            container = self.query_one('#command-pane-container', Vertical)
            for child in container.query(CommandPane):
                child.display = False

        # Update input prefix and placeholder.
        self._update_input_prefix(index)
        input_widget = self.query_one('#command-input', Input)
        input_widget.placeholder = self._get_input_placeholder(index)

    def _focus_sidebar(self) -> None:
        self.query_one('#command-list', ListView).focus()

    def _focus_terminal(self) -> None:
        pane_id = self._get_selected_pane_id()
        if pane_id is not None:
            try:
                self.query_one(f'#{pane_id}', CommandPane).focus()
            except NoMatches:
                pass

    def _select_prev_sub_command(self) -> None:
        select = self.query_one('#command-select', Select)
        if select.value is Select.BLANK:
            return
        current: int = select.value  # type: ignore[assignment]
        if current > 0:
            select.value = current - 1

    def _select_next_sub_command(self) -> None:
        select = self.query_one('#command-select', Select)
        if select.value is Select.BLANK:
            return
        current: int = select.value  # type: ignore[assignment]
        tab = self._tabs[self._active_tab]
        if current < len(tab.sub_commands) - 1:
            select.value = current + 1

    def _on_focused_changed(self, focused: Optional[Widget]) -> None:
        if focused is None:
            self._focus_sidebar()

    def on_key(self, event: events.Key) -> None:
        focused = self.screen.focused
        sidebar = self.query_one('#command-list', ListView)

        # Tab / Shift+Tab: cycle between sidebar and terminal only.
        if event.key in ('tab', 'shift+tab'):
            event.stop()
            event.prevent_default()
            if not isinstance(focused, Menu):
                if event.key == 'tab':
                    self._focus_terminal()
                else:
                    self._focus_sidebar()
            return

        # The following shortcuts only apply when the sidebar is focused.
        if focused is not sidebar:
            return

        if event.character == '!':
            event.stop()
            event.prevent_default()
            self.query_one('#command-input', ShellInput).focus()
            return

        if event.key == 'left':
            event.stop()
            event.prevent_default()
            self._select_prev_sub_command()
            return

        if event.key == 'right':
            event.stop()
            event.prevent_default()
            self._select_next_sub_command()
            return

        if event.character == '?':
            event.stop()
            event.prevent_default()
            self.push_screen(HelpModal())
            return

    def on_select_changed(self, event: Select.Changed) -> None:
        if event.select.id != 'command-select':
            return
        pane_id = self._get_selected_pane_id()
        if pane_id is not None:
            self._show_pane(pane_id)

    def _refresh_select_if_active(self, tab_index: int):
        if tab_index == self._active_tab:
            self._refresh_select()

    def on_command_pane_command_complete(
        self, _event: CommandPane.CommandComplete
    ) -> None:
        for tab_index, tab in enumerate(self._tabs):
            for sub in tab.sub_commands:
                if sub.status != CommandStatus.RUNNING:
                    continue
                try:
                    pane = self.query_one(f'#{sub.pane_id}', CommandPane)
                except NoMatches:
                    continue
                if pane.return_code is None:
                    continue

                if pane.return_code == 0:
                    sub.status = CommandStatus.SUCCESS
                    pane.border_subtitle = 'Done'
                else:
                    sub.status = CommandStatus.FAILED
                    pane.border_subtitle = f'Exit code: {pane.return_code}'

                self._update_sidebar(tab_index)
                self._refresh_select_if_active(tab_index)
                if sub.task_id is not None:
                    self._task_queue.notify_complete(sub.task_id)
                return

    def _queue_command_in_tab(self, tab_index: int, raw_command: str) -> SubCommand:
        tab = self._tabs[tab_index]
        display_name = (
            f'{tab.entry.prefix} {raw_command}'
            if tab.entry.prefix is not None
            else raw_command
        )
        sub = tab.add_sub_command_raw(name=display_name, raw_command=raw_command)

        # Mount the new pane.
        container = self.query_one('#command-pane-container', Vertical)
        pane = _AppCommandPane(id=sub.pane_id)
        pane.border_title = sub.name
        pane.display = False
        container.mount(pane)

        self._update_sidebar(tab_index)
        self._refresh_select_if_active(tab_index)

        task = self._task_queue.enqueue(sub.shell_command, terminal_id=tab_index)
        sub.task_id = task.task_id
        return sub

    def _show_latest_sub_command(self) -> None:
        """Refresh the select widget and show the latest sub-command pane."""
        tab = self._tabs[self._active_tab]
        self._refresh_select()
        if tab.sub_commands:
            select = self.query_one('#command-select', Select)
            select.value = len(tab.sub_commands) - 1
            self._show_pane(tab.sub_commands[-1].pane_id)

    def _submit_command(self, raw_input: str):
        sub = self._queue_command_in_tab(self._active_tab, raw_input)
        self._show_latest_sub_command()
        if sub.status == CommandStatus.PENDING:
            self.notify(
                f'Command queued in {self._tabs[self._active_tab].entry.display_name}'
            )

    def _submit_command_all(self, raw_input: str):
        queued = 0
        for i in range(len(self._tabs)):
            sub = self._queue_command_in_tab(i, raw_input)
            if sub.status == CommandStatus.PENDING:
                queued += 1
        self._show_latest_sub_command()
        if queued > 0:
            self.notify(f'Command queued in {queued} tab(s)')

    def _dismiss_menu(self) -> None:
        self._pending_command = None
        for menu in self.query(Menu):
            menu.remove()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id != 'command-input':
            return
        raw = event.value.strip()
        if not raw:
            return
        event.input.value = ''

        # Dismiss any existing menu first.
        self._dismiss_menu()

        self._pending_command = raw
        menu = Menu(
            [
                MenuItem('Run in this tab', 'run_this_tab', '1'),
                MenuItem('Run in all tabs', 'run_all_tabs', '2'),
                MenuItem('Run in selected tabs', 'run_selected_tabs', '3'),
            ],
        )
        input_container = self.query_one('#command-input-container', Horizontal)
        input_container.mount(menu)
        menu.focus()

    @on(Menu.OptionSelected)
    def _on_menu_selected(self, event: Menu.OptionSelected) -> None:
        event.stop()
        raw = self._pending_command
        self._pending_command = None
        event.menu.remove()

        if raw is not None and event.action == 'run_selected_tabs':
            tab_names = [tab.entry.display_name for tab in self._tabs]
            self.push_screen(
                TabSelectorModal(tab_names),
                callback=lambda indices: self._on_tabs_selected(raw, indices),
            )
            return

        if raw is not None:
            if event.action == 'run_this_tab':
                self._submit_command(raw)
            elif event.action == 'run_all_tabs':
                self._submit_command_all(raw)
        self._focus_sidebar()

    @on(Menu.Dismissed)
    def _on_menu_dismissed(self, event: Menu.Dismissed) -> None:
        event.stop()
        self._pending_command = None
        event.menu.remove()

        # Clear input and return to sidebar.
        self.query_one('#command-input', ShellInput).value = ''
        self._focus_sidebar()

    @on(ShellInput.Escaped)
    def _on_shell_input_escaped(self, event: ShellInput.Escaped) -> None:
        event.stop()
        self.query_one('#command-input', ShellInput).value = ''
        self._focus_sidebar()

    def _on_tabs_selected(self, raw: str, indices: Optional[List[int]]) -> None:
        if indices:
            self._submit_command_selected(raw, indices)
        self._focus_sidebar()

    def _submit_command_selected(self, raw_input: str, tab_indices: List[int]) -> None:
        queued = 0
        for i in tab_indices:
            if 0 <= i < len(self._tabs):
                sub = self._queue_command_in_tab(i, raw_input)
                if sub.status == CommandStatus.PENDING:
                    queued += 1
        if self._active_tab in tab_indices:
            self._show_latest_sub_command()
        if queued > 0:
            self.notify(f'Command queued in {queued} tab(s)')


def start_command_app(commands: List[CommandEntry], parallel: bool = False) -> None:
    app = rbxCommandApp(commands, parallel=parallel)
    app.run()


if __name__ == '__main__':
    start_command_app(
        [
            CommandEntry(argv=['echo', 'hello'], name='echo1'),
            CommandEntry(argv=['echo', 'world'], name='echo2'),
            CommandEntry(argv=['echo', 'foo'], name='echo3'),
        ]
    )



import difflib
import pathlib
from typing import Optional, Tuple

from rich.markdown import Markdown
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import RichLog


def compute_diff(file1: pathlib.Path, file2: pathlib.Path) -> str:
    lines1 = file1.read_text().splitlines(keepends=True)
    lines2 = file2.read_text().splitlines(keepends=True)
    return ''.join(difflib.ndiff(lines1, lines2))


class DiffBox(Widget, can_focus=False):
    paths: reactive[Optional[Tuple[pathlib.Path, pathlib.Path]]] = reactive(None)

    def __init__(self):
        super().__init__()

    def compose(self) -> ComposeResult:
        md = RichLog()
        md.border_title = 'Differ'
        yield md

    async def watch_paths(self, paths: Optional[Tuple[pathlib.Path, pathlib.Path]]):
        log = self.query_one(RichLog)
        log.clear()
        if paths is None:
            return
        file1, file2 = paths
        md = Markdown(
            f'```diff\n{compute_diff(file1, file2)}\n```', code_theme='monokai'
        )
        log.write(md)



import pathlib
import sys
from unittest.mock import patch

import pytest

from rbx.grading import steps
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    GradingLogsHolder,
    RunLogMetadata,
)


@pytest.fixture(autouse=True)
def clear_steps_cache():
    """Clear all cached functions in steps.py before each test."""
    # Clear all cached functions to ensure clean state between tests
    steps._complain_about_clang.cache_clear()  # noqa: SLF001
    steps._get_cxx_version_output.cache_clear()  # noqa: SLF001
    steps._maybe_complain_about_sanitization.cache_clear()  # noqa: SLF001
    steps._try_following_alias_for_exe.cache_clear()  # noqa: SLF001


class TestStepsRun:
    """Test the steps.run function."""

    async def test_run_simple_python_command(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a simple Python command that outputs to stdout."""
        # Setup input file
        source_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=source_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert result.time is not None
        assert result.memory is not None
        assert (cleandir / 'output.txt').exists()
        assert (cleandir / 'output.txt').read_text().strip() == 'Hello from Python!'
        assert artifacts.logs.run is not None
        assert artifacts.logs.run.exitcode == 0

    async def test_run_with_input_file(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with input file redirection."""
        # Setup input files
        script_file = testdata_path / 'steps_run_test' / 'simple_input.py'
        input_file = testdata_path / 'steps_run_test' / 'input.txt'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(src=script_file, dest=pathlib.Path('script.py')),
                GradingFileInput(src=input_file, dest=pathlib.Path('input.txt')),
            ]
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdin_file=pathlib.Path('input.txt'), stdout_file=pathlib.Path('output.txt')
        )
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        assert 'Received: test input line' in (cleandir / 'output.txt').read_text()

    async def test_run_creates_output_file(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command that creates an output file."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'create_file.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('created.txt'), dest=pathlib.Path('created.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        command = f'{sys.executable} script.py created.txt "test content"'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'created.txt').exists()
        assert (cleandir / 'created.txt').read_text() == 'test content'
        assert (cleandir / 'stdout.txt').exists()
        assert (
            'Created file created.txt with content: test content'
            in (cleandir / 'stdout.txt').read_text()
        )

    async def test_run_with_error_returns_error_exitcode(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that run returns exitcode with error when command fails."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'error_program.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py 42'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None  # run returns RunLog even on error
        assert result.exitcode == 42
        assert artifacts.logs.run is not None
        assert artifacts.logs.run.exitcode == 42

    async def test_run_handles_program_error_and_returns_sandbox_error(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """ProgramError should yield a RunLog with sandbox error status."""
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        from rbx.grading.judge.program import ProgramError

        with patch.object(sandbox, 'run', side_effect=ProgramError('boom')):
            result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitstatus == SandboxBase.EXIT_SANDBOX_ERROR
        assert result.exitcode == 1
        assert artifacts.logs.run is not None

    async def test_run_with_timeout(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with CPU timeout."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'busy_loop.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            timeout=500,  # 500ms timeout
        )
        command = f'{sys.executable} script.py 2.0'  # Busy loop for 2 seconds

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        # Should timeout due to CPU time limit
        assert result.exitstatus == SandboxBase.EXIT_TIMEOUT
        assert result.exitcode != 0

    async def test_run_with_wall_timeout(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with wall timeout."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'slow_program.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            wallclock_timeout=500,  # 500ms wall timeout
        )
        command = f'{sys.executable} script.py 2.0'  # Sleep for 2 seconds

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        # Should timeout due to wall time limit
        assert result.exitstatus == SandboxBase.EXIT_TIMEOUT_WALL
        assert result.exitcode != 0

    async def test_run_with_memory_limit(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with memory limit."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'memory_heavy_gradual.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            address_space=50,  # 50MB memory limit
        )
        command = f'{sys.executable} script.py 100'  # Try to allocate 100MB

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        # Should hit memory limit and be killed
        assert result.exitstatus == SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED
        assert result.exitcode != 0

    async def test_run_with_memory_limit_regression_test(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Regression test for memory limit detection bug.

        This test ensures that the get_memory_usage function correctly reports
        memory usage in bytes on all platforms. Previously, on macOS, the function
        incorrectly divided ru.ru_maxrss by 1024, causing memory limits to not
        be properly detected.

        The original memory_heavy.py script allocates memory quickly and exits,
        relying on the post-execution memory check (ru.ru_maxrss) rather than
        the runtime monitoring thread to detect memory limit violations.
        """
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'memory_heavy.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            address_space=50,  # 50MB memory limit
        )
        command = f'{sys.executable} script.py 100'  # Try to allocate 100MB

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        # Should hit memory limit and be flagged (may complete successfully but exceed limit)
        assert result.exitstatus == SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED
        # Memory usage should exceed the limit
        assert result.memory is not None
        assert result.memory > 50 * 1024 * 1024  # Should use more than 50MB

    async def test_run_with_metadata(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with metadata."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        metadata = RunLogMetadata(
            language='python', is_sanitized=False, timeLimit=1000, memoryLimit=512
        )
        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts, metadata)

        assert result is not None
        assert result.exitcode == 0
        assert result.metadata is not None
        assert result.metadata.language == 'python'
        assert result.metadata.timeLimit == 1000
        assert result.metadata.memoryLimit == 512
        assert not result.metadata.is_sanitized
        assert artifacts.logs.run is not None
        assert artifacts.logs.run.metadata is not None
        assert artifacts.logs.run.metadata.language == 'python'

    async def test_run_java_removes_memory_constraints(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that Java commands have memory constraints removed."""
        # Setup input file
        source_file = testdata_path / 'steps_run_test' / 'simple.java'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=source_file, dest=pathlib.Path('Simple.java'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            address_space=1024,  # Set memory constraint
        )
        command = 'java Simple'

        # Use mock to capture the params passed to sandbox.run
        with patch.object(sandbox, 'run', wraps=sandbox.run) as mock_run:
            result = await steps.run(command, params, sandbox, artifacts)

            assert result is not None
            # Original params should not be modified
            assert params.address_space == 1024
            # But the params passed to sandbox.run should have address_space removed
            mock_run.assert_called_once()
            call_args = mock_run.call_args
            captured_params = call_args[0][1]  # Second argument (params)
            assert captured_params.address_space is None

    async def test_run_with_optional_output_file(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with optional output file that doesn't exist."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('optional.txt'),
                    dest=pathlib.Path('optional.txt'),
                    optional=True,
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        assert not (
            cleandir / 'optional.txt'
        ).exists()  # Optional file doesn't exist, but that's OK

    async def test_run_with_executable_output(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command that produces an executable output."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'create_file.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('executable'),
                    dest=pathlib.Path('executable'),
                    executable=True,
                ),
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        command = f'{sys.executable} script.py executable "#!/bin/bash\\necho test"'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'executable').exists()
        # Check that the file has executable permissions
        assert (cleandir / 'executable').stat().st_mode & 0o111

    async def test_run_missing_required_output_returns_none(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that run returns None when required output file is missing."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('missing.txt'),
                    dest=pathlib.Path('missing.txt'),
                    optional=False,  # Required file
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is None  # Should return None due to missing required output

    async def test_run_with_maxlen_output(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with output size limit."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'create_file.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('large.txt'),
                    dest=pathlib.Path('large.txt'),
                    maxlen=10,  # Limit to 10 bytes
                ),
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        # Create a file with more than 10 bytes
        large_content = 'x' * 100
        command = f'{sys.executable} script.py large.txt "{large_content}"'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'large.txt').exists()
        # The file should be truncated to maxlen
        assert len((cleandir / 'large.txt').read_text()) <= 10

    async def test_run_with_touch_output(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with touch output file."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('touched.txt'),
                    dest=pathlib.Path('touched.txt'),
                    touch=True,
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        assert (cleandir / 'touched.txt').exists()
        # The touched file should be empty (touched before command runs)
        assert (cleandir / 'touched.txt').read_text() == ''

    async def test_run_processes_input_artifacts(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that run processes input artifacts correctly."""
        # Setup multiple input files
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        input_file = testdata_path / 'steps_run_test' / 'input.txt'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(src=script_file, dest=pathlib.Path('script.py')),
                GradingFileInput(src=input_file, dest=pathlib.Path('data.txt')),
            ]
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        # Both input files should have been processed into the sandbox

    async def test_run_sanitizer_warnings_detection(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that sanitizer warnings are detected when metadata indicates sanitization."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('stderr.txt'), dest=pathlib.Path('stderr.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        metadata = RunLogMetadata(language='cpp', is_sanitized=True)
        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            stderr_file=pathlib.Path('stderr.txt'),
        )
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts, metadata)

        assert result is not None
        assert result.exitcode == 0
        assert result.metadata is not None
        assert result.metadata.is_sanitized
        # The warnings flag should be set based on stderr content
        # (In this case, no actual sanitizer warnings, so should be False)
        assert not result.warnings

    async def test_run_logs_are_captured(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that run logs are properly captured in artifacts."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert artifacts.logs is not None
        assert artifacts.logs.run is not None
        assert artifacts.logs.run.exitcode == 0
        assert artifacts.logs.run.time is not None
        assert artifacts.logs.run.memory is not None
        assert artifacts.logs.run.sandbox is not None
        assert artifacts.logs.run.exitstatus is not None

    async def test_run_without_logs_holder(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that run works without a logs holder."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
            )
        )
        # No logs holder set

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        # artifacts.logs should remain None
        assert artifacts.logs is None

    async def test_run_with_actual_sanitizer_warnings(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that actual sanitizer warnings are detected."""
        # Setup input file that simulates sanitizer warnings
        script_file = testdata_path / 'steps_run_test' / 'sanitizer_warning.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('output.txt'), dest=pathlib.Path('output.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('stderr.txt'), dest=pathlib.Path('stderr.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        metadata = RunLogMetadata(language='cpp', is_sanitized=True)
        params = SandboxParams(
            stdout_file=pathlib.Path('output.txt'),
            stderr_file=pathlib.Path('stderr.txt'),
        )
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts, metadata)

        assert result is not None
        assert result.exitcode == 0
        assert result.metadata is not None
        assert result.metadata.is_sanitized
        # The warnings flag should be set based on stderr content
        assert result.warnings  # Should detect sanitizer warnings

    async def test_run_with_intermediate_output(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with intermediate output files."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'create_file.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('script.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('intermediate.txt'),
                    dest=pathlib.Path('intermediate.txt'),
                    intermediate=True,  # Intermediate file
                ),
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        command = f'{sys.executable} script.py intermediate.txt "intermediate content"'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'intermediate.txt').exists()
        assert (cleandir / 'intermediate.txt').read_text() == 'intermediate content'
        assert (cleandir / 'stdout.txt').exists()

    async def test_run_with_no_hash_tracking(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test running a command with hash tracking disabled."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'simple_output.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(
                src=script_file, dest=pathlib.Path('script.py'), hash=False
            )
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('output.txt'),
                dest=pathlib.Path('output.txt'),
                hash=False,  # Disable hash tracking
            )
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('output.txt'))
        command = f'{sys.executable} script.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'output.txt').exists()
        assert (cleandir / 'output.txt').read_text().strip() == 'Hello from Python!'

    async def test_run_with_glob_expansion(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that @glob: command expansion works correctly."""
        # Setup input files - script and files to be matched by glob
        script_file = testdata_path / 'steps_run_test' / 'glob_test.py'
        test1_file = testdata_path / 'steps_run_test' / 'test1.py'
        test2_file = testdata_path / 'steps_run_test' / 'test2.py'
        other_file = testdata_path / 'steps_run_test' / 'other.txt'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(src=script_file, dest=pathlib.Path('glob_test.py')),
                GradingFileInput(src=test1_file, dest=pathlib.Path('test1.py')),
                GradingFileInput(src=test2_file, dest=pathlib.Path('test2.py')),
                GradingFileInput(src=other_file, dest=pathlib.Path('other.txt')),
            ]
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('glob_output.txt'),
                    dest=pathlib.Path('glob_output.txt'),
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        # Use @glob: to expand *.py files
        command = f'{sys.executable} glob_test.py @glob:*.py'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'stdout.txt').exists()
        assert (cleandir / 'glob_output.txt').exists()

        # Check that glob expansion worked - should find test1.py and test2.py
        stdout_content = (cleandir / 'stdout.txt').read_text()
        glob_output_content = (cleandir / 'glob_output.txt').read_text()

        # Should have found 3 Python files (glob_test.py, test1.py, test2.py)
        assert 'Number of arguments: 3' in stdout_content

        # All Python files should be in the arguments
        assert 'glob_test.py' in glob_output_content
        assert 'test1.py' in glob_output_content
        assert 'test2.py' in glob_output_content

        # The .txt file should not be included
        assert 'other.txt' not in glob_output_content

        # Verify the arguments are properly quoted (shlex.quote behavior)
        # The order may vary, so just check that all files are present
        assert 'Count: 3' in glob_output_content

    async def test_run_with_glob_no_matches(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that @glob: works correctly when no files match the pattern."""
        # Setup input file
        script_file = testdata_path / 'steps_run_test' / 'glob_test.py'
        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.append(
            GradingFileInput(src=script_file, dest=pathlib.Path('glob_test.py'))
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('glob_output.txt'),
                    dest=pathlib.Path('glob_output.txt'),
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        # Use @glob: with a pattern that won't match any files
        command = f'{sys.executable} glob_test.py @glob:*.nonexistent'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'stdout.txt').exists()
        assert (cleandir / 'glob_output.txt').exists()

        # Should have found 0 files
        stdout_content = (cleandir / 'stdout.txt').read_text()
        glob_output_content = (cleandir / 'glob_output.txt').read_text()

        assert 'Number of arguments: 0' in stdout_content
        assert 'Count: 0' in glob_output_content
        assert 'Arguments: ' in glob_output_content  # Empty arguments

    async def test_run_with_mixed_glob_and_normal_args(
        self, sandbox: SandboxBase, cleandir: pathlib.Path, testdata_path: pathlib.Path
    ):
        """Test that @glob: works correctly mixed with normal arguments."""
        # Setup input files
        script_file = testdata_path / 'steps_run_test' / 'glob_test.py'
        test1_file = testdata_path / 'steps_run_test' / 'test1.py'
        test2_file = testdata_path / 'steps_run_test' / 'test2.py'

        artifacts = GradingArtifacts(root=cleandir)
        artifacts.inputs.extend(
            [
                GradingFileInput(src=script_file, dest=pathlib.Path('glob_test.py')),
                GradingFileInput(src=test1_file, dest=pathlib.Path('test1.py')),
                GradingFileInput(src=test2_file, dest=pathlib.Path('test2.py')),
            ]
        )
        artifacts.outputs.extend(
            [
                GradingFileOutput(
                    src=pathlib.Path('stdout.txt'), dest=pathlib.Path('stdout.txt')
                ),
                GradingFileOutput(
                    src=pathlib.Path('glob_output.txt'),
                    dest=pathlib.Path('glob_output.txt'),
                ),
            ]
        )
        artifacts.logs = GradingLogsHolder()

        params = SandboxParams(stdout_file=pathlib.Path('stdout.txt'))
        # Mix normal arguments with @glob: expansion
        command = f'{sys.executable} glob_test.py --flag @glob:*.py --end'

        result = await steps.run(command, params, sandbox, artifacts)

        assert result is not None
        assert result.exitcode == 0
        assert (cleandir / 'stdout.txt').exists()
        assert (cleandir / 'glob_output.txt').exists()

        # Should have found 5 arguments: --flag, glob_test.py, test1.py, test2.py, --end
        stdout_content = (cleandir / 'stdout.txt').read_text()
        glob_output_content = (cleandir / 'glob_output.txt').read_text()

        assert 'Number of arguments: 5' in stdout_content

        # All arguments should be present
        assert '--flag' in glob_output_content
        assert '--end' in glob_output_content
        assert 'glob_test.py' in glob_output_content
        assert 'test1.py' in glob_output_content
        assert 'test2.py' in glob_output_content



from unittest.mock import patch

import pytest
import typer

from rbx.box import naming
from rbx.box.schema import Package
from rbx.box.statements.schema import Statement


class TestGetTitle:
    def test_returns_pkg_title_for_lang_when_present(self):
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )

        assert naming.get_problem_title('en', pkg=pkg) == 'English Title'

    def test_falls_back_to_pkg_name_when_title_missing_lang(self):
        pkg = Package(name='base-name', timeLimit=1000, memoryLimit=256)

        assert naming.get_problem_title('en', pkg=pkg) == 'base-name'

    def test_statement_title_overrides_pkg_title(self):
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )
        statement = Statement(name='statement', language='en', title='Custom Title')

        assert (
            naming.get_problem_title('en', statement=statement, pkg=pkg)
            == 'Custom Title'
        )

    def test_statement_without_title_uses_pkg_title(self):
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )
        statement = Statement(name='statement', language='en')

        assert (
            naming.get_problem_title('en', statement=statement, pkg=pkg)
            == 'English Title'
        )

    def test_uses_problem_package_when_pkg_none(self):
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'pt': 'Título'},
        )

        with patch(
            'rbx.box.naming.package.find_problem_package_or_die', return_value=pkg
        ):
            assert naming.get_problem_title('pt') == 'Título'

    # New tests for updated functionality

    def test_lang_none_falls_back_to_pkg_name_by_default(self):
        """When lang=None and fallback_to_title=False, should return pkg.name"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )

        assert naming.get_problem_title(lang=None, pkg=pkg) == 'base-name'

    def test_lang_none_with_statement_title_uses_statement_title(self):
        """When lang=None but statement has title, should use statement title"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )
        statement = Statement(name='statement', language='en', title='Custom Title')

        assert (
            naming.get_problem_title(lang=None, statement=statement, pkg=pkg)
            == 'Custom Title'
        )

    def test_lang_none_with_statement_no_title_falls_back_to_pkg_name(self):
        """When lang=None and statement has no title, should fall back to pkg.name"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )
        statement = Statement(name='statement', language='en')

        assert (
            naming.get_problem_title(lang=None, statement=statement, pkg=pkg)
            == 'base-name'
        )

    def test_fallback_to_title_true_with_single_title_succeeds(self):
        """When fallback_to_title=True and pkg has exactly one title, should use that title"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )

        assert (
            naming.get_problem_title(lang=None, pkg=pkg, fallback_to_title=True)
            == 'English Title'
        )

    def test_fallback_to_title_true_with_multiple_titles_raises_error(self):
        """When fallback_to_title=True and pkg has multiple titles, should raise typer.Exit"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )

        with pytest.raises(typer.Exit):
            naming.get_problem_title(lang=None, pkg=pkg, fallback_to_title=True)

    def test_fallback_to_title_true_with_no_titles_fallback_to_name(self):
        """When fallback_to_title=True but pkg has no titles, should raise error since len(titles) != 1"""
        pkg = Package(name='base-name', timeLimit=1000, memoryLimit=256)

        assert (
            naming.get_problem_title(lang=None, pkg=pkg, fallback_to_title=True)
            == 'base-name'
        )

    def test_fallback_to_title_true_with_statement_title_uses_statement_title(self):
        """When fallback_to_title=True but statement has title, should prioritize statement title"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )
        statement = Statement(name='statement', language='en', title='Custom Title')

        assert (
            naming.get_problem_title(
                lang=None, statement=statement, pkg=pkg, fallback_to_title=True
            )
            == 'Custom Title'
        )

    def test_fallback_to_title_false_with_multiple_titles_falls_back_to_pkg_name(self):
        """When fallback_to_title=False and no specific title found, should use pkg.name regardless of available titles"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )

        assert (
            naming.get_problem_title(lang=None, pkg=pkg, fallback_to_title=False)
            == 'base-name'
        )

    def test_lang_specified_but_missing_falls_back_correctly(self):
        """When lang is specified but not in titles, fallback behavior should work correctly"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title'},
        )

        # With fallback_to_title=False (default)
        assert (
            naming.get_problem_title(lang='fr', pkg=pkg, fallback_to_title=False)
            == 'base-name'
        )

        # With fallback_to_title=True and single title
        assert (
            naming.get_problem_title(lang='fr', pkg=pkg, fallback_to_title=True)
            == 'English Title'
        )

    def test_lang_specified_but_missing_with_multiple_titles_raises_error(self):
        """When lang is specified but not in titles, and fallback_to_title=True with multiple titles, should raise error"""
        pkg = Package(
            name='base-name',
            timeLimit=1000,
            memoryLimit=256,
            titles={'en': 'English Title', 'pt': 'Título'},
        )

        with pytest.raises(typer.Exit):
            naming.get_problem_title(lang='fr', pkg=pkg, fallback_to_title=True)



import os
import pathlib

import pytest
import typer

from rbx.box.statements.statement_utils import get_relative_assets


class TestGetRelativeAssets:
    """Test get_relative_assets function."""

    def test_basic_asset_resolution(self, chdir_tmp_path):
        """Test basic asset file resolution."""
        # Create test files in the working directory
        asset_file = chdir_tmp_path / 'test.png'
        asset_file.write_text('fake image content')

        assets = get_relative_assets(chdir_tmp_path, ['test.png'])

        assert len(assets) == 1
        abs_path, rel_path = assets[0]
        assert abs_path.name == 'test.png'
        assert rel_path == pathlib.Path('test.png')

    def test_glob_pattern_asset_resolution(self, chdir_tmp_path):
        """Test asset resolution with glob patterns."""
        # Create multiple test files
        for i in range(3):
            (chdir_tmp_path / f'image{i}.png').write_text(f'content {i}')

        # Use real glob implementation with actual files
        assets = get_relative_assets(chdir_tmp_path, ['*.png'])

        assert len(assets) == 3
        png_files = [rel_path.name for _, rel_path in assets]
        assert 'image0.png' in png_files
        assert 'image1.png' in png_files
        assert 'image2.png' in png_files

    def test_nonexistent_asset_raises_exit(self, chdir_tmp_path):
        """Test that non-existent asset without glob raises typer.Exit."""
        with pytest.raises(typer.Exit):
            get_relative_assets(chdir_tmp_path, ['nonexistent.png'])

    def test_asset_outside_relative_path_raises_exit(self, tmp_path):
        """Test that asset outside relative path raises typer.Exit."""
        # Create asset outside the base directory
        outside_dir = tmp_path.parent / 'outside'
        outside_dir.mkdir()
        outside_asset = outside_dir / 'outside.txt'
        outside_asset.write_text('outside content')

        base_dir = tmp_path / 'base'
        base_dir.mkdir()

        # Change to base directory for this test
        original_cwd = os.getcwd()
        try:
            os.chdir(base_dir)
            with pytest.raises(typer.Exit):
                get_relative_assets(base_dir, [str(outside_asset)])
        finally:
            os.chdir(original_cwd)

    def test_root_parameter_resolution(self, tmp_path):
        """Test asset resolution using the root parameter."""
        # Structure:
        # tmp_path/
        #   project/  <- relative_to
        #     assets/
        #       image.png
        #       data1.txt
        #       data2.txt
        #   other/    <- CWD

        project_dir = tmp_path / 'project'
        project_dir.mkdir()
        assets_dir = project_dir / 'assets'
        assets_dir.mkdir()

        (assets_dir / 'image.png').write_text('image')
        (assets_dir / 'data1.txt').write_text('data1')
        (assets_dir / 'data2.txt').write_text('data2')

        other_dir = tmp_path / 'other'
        other_dir.mkdir()

        original_cwd = os.getcwd()
        try:
            os.chdir(other_dir)

            # Case 1: strict file match via root
            # We request 'assets/image.png'.
            # specific file check in CWD fails.
            # root.glob('assets/image.png') finds it in project_dir.
            assets = get_relative_assets(
                relative_to=project_dir,
                assets=['assets/image.png'],
                root=project_dir,
            )

            assert len(assets) == 1
            assert assets[0][0].name == 'image.png'
            # Expected relative path matches what was requested if it was relative to relative_to
            assert assets[0][1] == pathlib.Path('assets/image.png')

            # Case 2: Glob match via root
            assets_glob = get_relative_assets(
                relative_to=project_dir,
                assets=['assets/*.txt'],
                root=project_dir,
            )
            assert len(assets_glob) == 2
            names = sorted([p[0].name for p in assets_glob])
            assert names == ['data1.txt', 'data2.txt']

            # Case 3: Not found in root raises Exit
            with pytest.raises(typer.Exit):
                get_relative_assets(
                    relative_to=project_dir,
                    assets=['nonexistent.png'],
                    root=project_dir,
                )

        finally:
            # Always clean up cwd
            os.chdir(original_cwd)

    def test_root_resolution_error_when_asset_not_relative_to_statement(self, tmp_path):
        """Test error when asset exists in root but isn't relative to statement."""
        # tmp_path/
        #   root/
        #     project/  <- relative_to
        #     asset.png <- in root, but outside project
        root_dir = tmp_path / 'root'
        root_dir.mkdir()
        project_dir = root_dir / 'project'
        project_dir.mkdir()

        (root_dir / 'asset.png').write_text('asset')

        original_cwd = os.getcwd()
        try:
            os.chdir(tmp_path)
            # We ask for 'asset.png' (in root). It exists.
            # relative_to is 'root/project'.
            # 'root/asset.png' is NOT relative to 'root/project'.
            with pytest.raises(typer.Exit):
                get_relative_assets(
                    relative_to=project_dir,
                    assets=['asset.png'],
                    root=root_dir,
                )
        finally:
            os.chdir(original_cwd)

    def test_root_resolution_with_absolute_relative_to(self, tmp_path):
        """Test root resolution when relative_to is passed as absolute path."""
        # tmp_path/
        #   root/
        #     project/  <- relative_to
        #       asset.png
        root_dir = tmp_path / 'root'
        root_dir.mkdir()
        project_dir = root_dir / 'project'
        project_dir.mkdir()

        (project_dir / 'asset.png').write_text('asset')

        original_cwd = os.getcwd()
        try:
            os.chdir(tmp_path)
            # relative_to is absolute path to project_dir
            # assets must be relative to ROOT. So 'project/asset.png'.
            assets = get_relative_assets(
                relative_to=project_dir.resolve(),
                assets=['project/asset.png'],
                root=root_dir,
            )

            assert len(assets) == 1
            assert assets[0][0].name == 'asset.png'
            # path relative to relative_to:
            # /tmp/.../root/project/asset.png relative to /tmp/.../root/project is 'asset.png'
            assert assets[0][1] == pathlib.Path('asset.png')

        finally:
            os.chdir(original_cwd)


@pytest.fixture
def chdir_tmp_path(tmp_path):
    """Fixture to change to tmp_path directory and restore original directory after test."""
    original_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(original_cwd)
